""" a series of converter to transform L{gluon.dal} objects into Ext JS widgets.

@author: R. Le Gac

"""
import gluon.dal
import re

from extjs import *
from gluon import current
from helper import (encode_field,
                    get_field_validators,
                    get_foreign_field,
                    get_set_field,
                    get_store_id,
                    is_foreign_field,
                    is_set_field)
from mapper import map_default

# Associated gluon.dal.field type with an ExtJS widget
# The dictionary contains configuration parameters of the Ext JS widget
FTYPE_TO_XTYPE = {'boolean':   (CheckBox, {}),\
                  'date':      (DateField, {'format': 'Y-m-d'}),\
                  'datetime':  (DateField, {'format': 'Y-m-d H:i:s'}),\
                  'double':    (NumberField, {'decimalPrecision':2,
                                              'decimalSeparator': '.'}),\
                  'id':        (TextField, {}),\
                  'integer':   (NumberField, {}),\
                  'password':  (TextField, {'inputType': 'password'}),\
                  'reference': (ComboBox, {}),\
                  'string':    (TextField, {}),\
                  'text':      (TextArea, {}),\
                  'time':      (TimeField, {}),\
                  'upload':    (TextField, {'inputType': 'file'})}


# constant defining a ExtJS store
ROOT = 'records'
SUCCESS= 'success'
TOTAL = 'count'


def _to_field(field, **kwargs):
    """Build the configurator L{Field} associated to the database C{field}.
    The conversion takes into account the L{FieldsModifier} instructions
    but does not handle C{CompositeField}.
    
    @param field:
    @type field: gluon.dal.Field
    
    @keyword kwargs: any Ext JS configuration parameters of the 
    C{Ext.form.Field} class. They are applied in the following order:
    constructor, modifiers and keyword arguments.
    
    @rtype: Field
    
    @note: The return configurator is a L{Field}. However it can be a 
    L{ComboBox}, a L{DateField}, a L{TextArea}, ... 
    depending on the C{type} of the database field.
    The association is defined in the dictionary L{FTYPE_TO_XTYPE}.
     
    @attention: It is recommended to used the more general method L{to_field}
    which also understands C{CompositeField}.
    
    """
    T = current.T
    
    fieldname = field.name
    tablename = field.tablename

    f_name = encode_field(tablename, fieldname)
                
    # foreign key
    if is_foreign_field(field):
            
        k_tablename, k_fieldname, k_key = get_foreign_field(field)
        
        cfg = ComboBox(name=f_name,
                       hiddenName=f_name,
                       displayField=encode_field(k_tablename, k_fieldname),
                       valueField=encode_field(k_tablename, k_key),
                       store=get_store_id(k_tablename))

    # set field
    elif is_set_field(field):
        cfg = ComboBox(name=f_name,
                       store=get_set_field(field))
        
    # other types of fields
    else:
        cls, di = FTYPE_TO_XTYPE[field.type] 
        cfg = cls(name=f_name, **di)

    # field label translate in the local languqge
    cfg["fieldLabel"] = str(T(field.label))

    # default value and not null
    # NOTE: the entryForm doesn't work when both default and
    # notnull condition are defined.
    if field.default:
        cfg["value"] = str(field.default)
    elif field.notnull:
        cfg["allowBlank"] = False

    # read only field
    if field.readable and not field.writable:
        cfg["readOnly"] = True

    # tool tip
    if field.comment:
        cfg["tipText"] = field.comment
     
    # validator constraints
    cfg.update(get_field_validators(field))
    
    # hide primary key
    # hide field which are not readable and not writable
    # hide field request by the form modifier
    hidden = field.name == "id"
    hidden = hidden or ((not field.readable) and (not field.writable))
    
    modifier_forms = PluginManager('dbui').dbui.modifier_forms
    if tablename in modifier_forms:
        if fieldname in modifier_forms[tablename].hidden_fields:
            hidden = True
            
    if hidden:
        cfg["hidden"] = True
        cfg["hideLabel"] = True
        cfg["readOnly"] = True

    # configuration options set by the field_modifers
    modifier_fields = PluginManager('dbui').dbui.modifier_fields
    if tablename in modifier_fields:
        if fieldname in modifier_fields[tablename].extjs_fields:
            cfg.update(modifier_fields[tablename].extjs_fields[fieldname])

    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg


def to_field(field, composite=True, **kwargs):
    """Build the configurator L{Field} for the database field.
    The conversion handles L{CompositeField} and takes into account
    the C{FieldsModifier} instructions.

    @note: The return configurator is a L{Field}. However it can be a 
    L{ComboBox}, a L{DateField}, a L{TextArea}, ... 
    depending on the C{type} of the database field.
    The association is defined in the dictionary L{FTYPE_TO_XTYPE}.
    
    @note: The database fields can be organized in L{CompositeField}
    in the application model using the modifiers L{FieldsModifier}.

    @type field: gluon.dal.Field
    @param field:
     
    @type composite: bool
    @param composite: C{CompositeField} is ignored when C{False}.
     
    @param kwargs: any Ext JS configuration parameter of the C{Ext.form.Field}
    class. They are applied in the following order:
    constructor, modifiers and keyword arguments.
    
    @rtype: Field or C{None}
    
    @note: Return None when composite is C{True} and when
    the field is consumed by a C{CompositeField}.
    
    """
    table = field.table
    tablename = field.tablename
    modifier_fields = PluginManager('dbui').dbui.modifier_fields

    # do we have composite field for this table ?
    composite_fields = None
    if tablename in modifier_fields:
        modifier_field = modifier_fields[tablename]
        composite_fields = modifier_field.composite_fields

    # main field of the composite field
    # it will consume the embedded field too
    if composite and composite_fields and field.name in composite_fields.main:
            
        cfg = CompositeField()

        for fieldname in composite_fields[field.name].fields:
            cfg.append_items(_to_field(table[fieldname]))
        
        # use configuration options from the modifier
        cfg.update(composite_fields[field.name].extjs)
    
    # an embedded field already used in a composite field
    elif composite and composite_fields and field.name in composite_fields.others:
        cfg = None
    
    # a basic field
    else:
        cfg = _to_field(field)

    # configuration options from keyword arguments
    if cfg:
        cfg.update(kwargs)
    
    return cfg

    
def to_fields(table):
    """Build the list of configurators L{Field} for each field 
    of the database C{table}. 

    @note: The return configurator is a L{Field}. However it can be a 
    L{ComboBox}, a L{DateField}, a L{TextArea}, ... 
    depending on the C{type} of the database field.
    The association is defined in the dictionary L{FTYPE_TO_XTYPE}.
    
    @note: The function handles L{FieldSet} and L{CompositeField}
    as well as the L{FieldsModifier} instructions.
    The C{FieldSet} and C{CompositeField} can be defined in the application
    model using the modifier L{FormModifier} and the L{FieldsModifier} 
    respectively.

    @type table: gluon.dal.Table
    @rtype: list
     
    """
    li = []
    tablename = table._tablename
    modifier_forms =  PluginManager('dbui').dbui.modifier_forms

    # do we have FieldSets
    field_sets, modifier_form = None, None
    if tablename in modifier_forms:
        modifier_form = modifier_forms[tablename]
        if modifier_form and modifier_form.field_sets:
            field_sets = modifier_form.field_sets
            
    # Table with Ext.form.FieldSet and/or Ext.form.CompositeFields 
    # and/or Ext.form.Field and/or Spacer

    # NOTE: the field id is add in order to run this form
    # in all context, create, update, destroy.
    # It is add to the latest fieldset.
    id_is_used = False 

    if field_sets:
        for fieldset in field_sets:
            
            cfg = FieldSet()            
            for fieldname in fieldset.fields:
                        
                if isinstance(fieldname, Base):
                    cfg.append_items(fieldname)
                    continue

                if fieldname == 'id':
                    id_is_used = True

                field = table[fieldname]
                item = to_field(field)
                if item:
                    cfg.append_items(item)
        
            # fieldset configuration options ser by the modifier
            cfg.update(fieldset.extjs)
            li.append(cfg)
            
        # append the field Id the last field set
        if not id_is_used:
            li[-1].append_items(to_field(table['id']))

    # Table with Ext.form.CompositeFields and/or Ext.form.Fields
    else:
        for field in table:
            cfg = to_field(field)
            if cfg:
                li.append(cfg)
    
    # map the list of fields/fieldSets on Ext.form.formPanel
    mapper = map_default
    if modifier_form and modifier_form.mapper:
        mapper = modifier_form.mapper

    li = mapper(li)

    return li


def to_formPanel(table, **kwargs):
    """Build the configurator L{FormPanel} for the database C{table}. 
    The conversion takes into account the L{FormModifier} instructions.

    @type table: gluon.dal.Table
    @param table:
    
    @param kwargs: any Ext JS configuration parameter of the 
    C{Ext.form.FormPanel} class. They are 
    applied in the following order: constructor, modifiers 
    and keyword arguments.
        
    @rtype: FormPanel

    """
    tablename = table._tablename

    fields = to_fields(table)
    cfg = FormPanel(items=fields)

    # add a store for non dummy database        
    if table._db._uri:
        cfg['store'] = get_store_id(tablename)
    
    # configuration options from form modifier
    modifier_forms = PluginManager('dbui').dbui.modifier_forms
    if tablename in modifier_forms:
        cfg.update(modifier_forms[tablename].extjs)
        
    # configuration options form the keyword arguments
    cfg.update(kwargs)
    
    return cfg 


def to_gridColumn(field, **kwargs):
    """Build the configurator L{GridColumn} for the database C{field}. 
    The conversion takes into account the L{GridModifier} instructions.

    @type field: gluon.dal.Field
    @param field:
    
    @param kwargs: any Ext JS configuration parameter of the 
    C{Ext.grid.Column} class. 
    They are applied in the following order: constructor, modifiers 
    and keyword arguments.
        
    @rtype: GridColumn
            
    """
    T = current.T
    
    fieldname = field.name
    tablename = field.tablename
    
    cfg = GridColumn(header=str(T(field.label)),
                     sortable=True)
    
    # name of the field in the data store
    # replace foreign key by the pointing column
    if is_foreign_field(field):
        k_tablename, k_fieldname, k_key = get_foreign_field(field)
        index = encode_field(k_tablename, k_fieldname)

    else:
        index  = encode_field(tablename, fieldname)

    cfg["dataIndex"] = index
        
    # hide the primary key
    if fieldname == "id":
        cfg["hidden"] = True

    # hide field which are not readable and not writable
    if (not field.readable) and (not field.writable):
        cfg["hidden"] = True

    # Hide fields request by the grid modifiers
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        if fieldname in modifier_grids[tablename].hidden_columns:
            cfg["hidden"] = True
    
    # configuration options from the grid modifier
    if tablename in modifier_grids:
        if fieldname in modifier_grids[tablename].configure_columns:
            cfg.update(modifier_grids[tablename].configure_columns[fieldname])
    
    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg


def to_gridColumnModel(table):
    """Build the configurator L{GridColumnModel} for the database C{table}. 
    The conversion takes into account the L{GridModifier} instructions.

    @type table: gluon.dal.Table
    @param table:
    
    @rtype: GridColumnModel
            
    """
    delete_columns = []
    template_columns = []
    tablename = table._tablename

    # get modifier requirements
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        delete_columns = modifier_grids[tablename].delete_columns
        template_columns = modifier_grids[tablename].template_columns

    # standard column
    cfg = GridColumnModel()  
    for field in table:
        fieldname = field.name
        if fieldname not in delete_columns:
            cfg.append(to_gridColumn(field))
    
    # template column
    for tpl in template_columns:
        col = GridTemplateColumn(sortable=True)
        col.update(tpl.extjs)
        cfg.insert(tpl.position, col)

    # row numbering in the first column
    if modifier_grids[tablename].row_numbering:
        cfg.insert(0, GridRowNumberer())
        
    return cfg


def to_gridFilter(table, **kwargs):
    """Build the configurator L{GridFilter} for thedatabase C{table}.
    The GridFilter is parametrized using the L{GridModifier}.
    
    @type table: gluon.dal.Table
    @param table:
    
    @keyword kwargs: any of the Ext JS configuration parameters of the
    C{App.grid.GridFilter} class. They are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    @rtype: GridFilter

    @note: Return an empty dictionary if the GridFilter is
    not defined in the application model.
    
    @note: A filter rule is defined in the model via the L{GridModifier}.
    The rule is a C{tuple} containing 3 strings 
    C{("field1", "operator", "comment")}
    where C{field1} is the name of a field belonging to the C{table}.
    A more elaborate filter rule allows to filter on any foreign field using
    the syntax is C{("table2.field1", "operator", "comments")}.
    Valid operators are defined in the method L{dbsvc.DbSvc._encode_query}.
    
    @note: Rules are store in the modifier section C{grid_filters}.
        
    """    
    T = current.T
    tablename = table._tablename
     
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename not in modifier_grids:
        return {}
        
    grid_filters = modifier_grids[tablename].grid_filters
    if not grid_filters:
        return {}

    if not grid_filters.filters:
        return {}
    
    di = GridFilter(title=T('Filter %s' % tablename))
    
    for i in range(len(grid_filters.filters)):
        
        fieldname, operator, comment = grid_filters.filters[i]
        
        # filter rule dealing with any foreign field
        m = re.match(r"(\w+)\.(\w+)", fieldname)
        if m:
            ktablename = m.group(1)
            kfieldname = m.group(2)
            field = table._db[ktablename][kfieldname]
            name = '[%s.%s]' % (ktablename, kfieldname)
            
        # basic filter rule
        else:
            field = table[fieldname]
            name = '[%s.%s]' % (tablename, fieldname)
        
        # get the standard configuration for the field
        # ignore composite field defined for form
        cfg = to_field(field, composite=False)
        
        # remove field constraints making no sense for a filter
        if 'allowBlank' in cfg:
            cfg["allowBlank"] = True

        if 'readOnly' in cfg:
            cfg['readOnly'] = False
                        
        if 'value' in cfg:
            del cfg['value']

        # replace all textarea by textfield
        if cfg['xtype'] == 'xtextarea':
            cfg['xtype'] = 'xtextfield'

        # replace all linkedcombobox by combobox
        if cfg['xtype'] == 'xlinkedcombobox':
            cfg['xtype'] = 'xcombobox'

        # empty text for combobox
        if cfg['xtype'] == 'xcombobox':
            cfg['emptyText']= T('select...')
        
        # prepare the name for the where close [table.field]
        cfg['name'] = name    
            
        # store information to customise the filter
        cfg['filterComment'] = comment  
        cfg['filterOperator'] = operator
        cfg['filterType'] = field.type
                    
        # copy the comment to be used as a tool tip
        # when the plugin pFormToolTip is used
        cfg['tipText'] = comment
        
        # field configuration option
        cfg.update(grid_filters.extjs_filters[i])
        
        di.append_items(cfg)
    
    # configuration options from the grid modifier
    di.update(grid_filters.extjs)
    
    # configuration option from keyword arguments
    di.update(kwargs)
    
    return di
    

def to_gridPanel(table, **kwargs):
    """Build the configurator L{GridPanel} for the database C{table}.
    The conversion takes into account the L{GridModifier} instructions.

    @type table: gluon.dal.Table
    @param table:
    
    @keyword kwargs: any of the Ext JS configuration parameters of the
    C{App.grid.GridPanel} class. They are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    @rtype: GridPanel

    """
    tablename = table._tablename

    cfg = GridPanel(columns=to_gridColumnModel(table),
                    frame=False,
                    viewConfig={'forceFit': True})
    
    # add a store for non dummy database        
    if table._db._uri:
        cfg['store'] = get_store_id(tablename)
    
    # configuration option from the grid modifier
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        cfg.update(modifier_grids[tablename].extjs)

    # configuration option from the keyword arguments
    cfg.update(kwargs)
    
    # grid with filter
    filter = to_gridFilter(table)
    if filter:
        cfg = GridWithFilter(panelCfg=dict(cfg),
                             selectorCfg=filter)
    
    return cfg


def to_jsonstore(table, **kwargs):
    """Build the configurator L{DirectStore} for the database C{table}.
    The conversion takes into account the L{StoreModifier} instructions.

    @type table: gluon.dal.Table
    @param table:
    
    @keyword kwargs: any of the Ext JS configuration parameters of the
    C{App.data.DirectStore} class. They are applied in the following order:
    constructor, modifiers, keyword arguments.
    
    @rtype: DirectStore

    """
    db = table._db
    tablename = table._tablename

    # NOTE: the configuration option baseParams defined the
    # transactions parameters seen by the server
    # This dictionary contains 3 keys:
    #    tablename: name of the table in the database
    #    dbFields: list of fields which should be red
    #    where: define inner join, mainly to resolve foreign key
    base_params = {'tableName': tablename, 'dbFields': []}
    
    cfg = DirectStore(autoLoad=False,
                      autoSave=True,
                      baseParams=base_params,
                      fields=[],
                      idProperty=encode_field(tablename,'id'),
                      root=ROOT,
                      successProperty=SUCCESS,
                      storeId=get_store_id(tablename),
                      totalProperty=TOTAL)

    for field in table:        
        fieldname = field.name

        # the field of the table
        # encode field name and define default value
        field_cfg = {'name': encode_field(tablename, fieldname)}
        if field.default:
            field_cfg['defaultValue'] = str(field.default)
        
        base_params['dbFields'].append([tablename, fieldname])
        cfg['fields'].append(field_cfg)
        
        # NOTE: the store contains all fields of the table.
        # It also contains the pointing values for foreign keys
        # Therefore the store has the the id and the value for 
        # each foreign key. This technique is use to resolve foreign key
        # in form (combobox) and in grid.
        if is_foreign_field(field):
            
            if 'where' not in base_params: base_params['where'] = []
            k_tablename, k_fieldname, k_key = get_foreign_field(field)
            
            # encode the field name
            k_field_cfg = {'name': encode_field(k_tablename, k_fieldname)}
            base_params['dbFields'].append([k_tablename, k_fieldname])
            cfg['fields'].append(k_field_cfg)
            
            # retrieve the default value for the foreign key
            if field.default:
                row = db(db[k_tablename].id == field.default).select().first()
                if row:
                    k_field_cfg['defaultValue'] = row[k_fieldname]
            
            # define the where close
            nt = (tablename, fieldname, k_tablename, k_key)
            base_params['where'].append("[%s.%s] == [%s.%s]" % nt)
    
    # configuration options from the store modifier
    # namely the order by directive
    modifier_stores = PluginManager('dbui').dbui.modifier_stores
    if tablename in modifier_stores:
        
        base_params['orderby'] = []
        for el in modifier_stores[tablename].orderby:

            if isinstance(el, gluon.dal.Field):
                field, dir = el, 'ASC'
                
            elif isinstance(el, gluon.dal.Expression):
                field, dir = el.first, 'DESC'

            base_params['orderby'].append([field.tablename, field.name, dir])

        # configuration option from the grid modifier
        cfg.update(modifier_stores[tablename].extjs)
            
            
    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg


def to_panelWithUrlSelector(table, selectorTitle='Select', **kwargs):
    """Build the configurator L{PanelWithUrlSelector}
    where the selector is a form derived from the database C{table}.

    @type table: gluon.dal.Table
    @param table:

    @type selectorTitle: str
    @param selectorTitle: the title of the FieldSet encapsulating
    the selector fields.
    
    @keyword baseUrl: The selected values are send to a controller 
    defined by the baseUrl.
    Values are processed and results will be published in the panel.
    It is mandatory to specified the baseUrl.
     
    @keyword kwargs: any of the Ext JS configuration parameters of the
    C{App.PanelWithUrlSelector} class.
    
    @rtype: PanelWithUrlSelector

    @raise BaseException: when baseUrl is not defined.
    
    """
    if 'baseUrl' not in kwargs:
        raise BaseException('The keyword argument basseUrl is missing.')

    # build the selector from the table
    fields = to_fields(table)
    selector = FieldSet(items=fields, 
                        plugins=['pFormToolTip'],
                        title=current.T(selectorTitle))

    # build the configuration for the PanelWithUrlSelector
    default = dict(isMathJax=False,
                   panelCfg=Panel(),
                   selectorCfg=selector,
                   selectorCollapsible=False)

    default.update(kwargs)
    cfg = PanelWithUrlSelector(**default)
    
    return cfg


def to_tree(node):
    """Build the list of L{Node}.
    The Node structure is defined in the application model using 
    L{ViewportModifier}.
    
    @type node: str
    @param node: is required by ExtJS 3.4 but it is not
    used here.
    
    @rtype: list
    
    """
    cfg = []
    
    modifier_viewports = PluginManager('dbui').dbui.modifier_viewports

    if modifier_viewports:
        nav_nodes = modifier_viewports.nodes
        for node in nav_nodes:
            cfg.append(node.get_node())

    return cfg


def to_viewport(**kwargs):
    """Build the configuration options dictionary for the C{Ext.Viewport}.
    The conversion takes into account the L{ViewportModifier} instructions.
    
    @keyword kwargs: any of the Ext JS configuration parameters of the
    C{App.Viewport} class.
    
    @rtype: dict
    
    """
    cfg = {}
    
    modifier_viewports = PluginManager('dbui').dbui.modifier_viewports

    if modifier_viewports:
        cfg = modifier_viewports.extjs
        
        if modifier_viewports.default:
            cfg['defaultTabPath'] = modifier_viewports.default
        
    # configuration options from the keyword arguments
    cfg.update(kwargs)
    
    return cfg