""" the database service.

@author: R. Le Gac

"""

import re

from basesvc import BaseSvc
from converter import ROOT, SUCCESS, TOTAL
from gluon import current
from gluon.contrib import simplejson as json
from gluon.storage import Storage
from helper import (decode_field,
                    encode_field, 
                    encode_validator_errors,
                    get_all_tables,
                    get_foreign_field,
                    is_foreign_field,
                    rows_serializer)

CALLBACK_ERRORS = "_callback_errors"

FIELD_NOT_IN_DB = "the field '%s.%s' doesn't exist in the database."
KEYWORD_MISSING = "The keyword '%s' is missing."
RECORD_NOT_IN_DB = "The record '%s' doesn't exist in the table %s."
RECORDS_DELETED = "Record(s) %s is deleted."
RECORDS_INSERTED = "Record(s) is inserted with id(s) %s."
RECORDS_UPDATED = "Record(s) %s is updated."
REJECT_BY_CALLBACK = "Operation rejected by the table callbacks."
TABLE_NOT_IN_DB = "The table '%s' doesn't exist in the database."
TO_MANY_RECORDS = "To many records !!!"


# transform bracket "[table1.field1]" as "db.table1.field1"
REG_BRACKET = re.compile(r'\[(\w+).(\w+)\]')

# True when the string match the pattern: 
# "db.publications.id_tablename == db.tablename.id
# The first group contains "tablename"
REG_IS_JOIN = re.compile(r'db\.\w+\.id_(?P<ref>\w+) == db\.(?P=ref)\.id')



class DbSvcException(BaseException): pass


class DbSvc(BaseSvc):
    """Interface the database and the client request.
    It is the glue between wep2py and Ext JS library.
    This service is the companion of L{DirectSvc} which implements 
    the C{Ext.Direct} protocol on the server side. 
    This protocol exposes methods of the C{DbSvc} service directly
    on the client-side.

    The main methods allow to:

        - C{create} a new record in a table.
        
        - C{read} the content of a a table
          resolving foreign key.

        - C{update} the record of a table.

        - C{destroy} the record of a table.

    These methods receive transaction data dictionary, C{arg}:

        - C{tableName}
          the name of the table in the database
            
        - C{dbFields}
            
            - C{[["table1", "field1"], ... ]}
            
            - The list of database fields of the table.
            
            - Include foreign keys and reference fields.

        - C{records} (optional)
            
              - C{[{"Table1Field1": val1,...}]}
              
              - A list of dictionaries containing key, value pairs for fields
                to be created or updated
              
              - Includes foreign key and reference field values.
              
              - Contains the C{id} of the record to be updated.
              
              - The keys are encoded as C{Table1Field1}.
          
              - The C{records} can also be a list containing the C{id} of the
                records to be deleted. 

    Each method return a dictionary:
        
        - C{{success: True, records:[{blabla}, ..], msg: 'blalbla'}}
    
        - The keyword C{success} is always present. It is mandatory for 
          the C{FormPanel} widget of the Ext JS library, but not strictly 
          required since failure can be propagated through HTTP errors.
    
        - In the records dictionary database field are encoded as C{TableField}.

        - If the action failed the C{DbSvcException} is raised.
          The exception can be catch to generate an HTTP error.

    """

    def _check_request(self, arg, keywords=('tableName', 'records')):
        """Check that the transaction data dictionary C{arg} 
        contains the C{keywords}.
    
        @type arg: dict
        @param arg: transaction data

            - C{tableName}
              the name of the table in the database
                
            - C{dbFields}
                
                - C{[["table1", "field1"], ... ]}
                
                - The list of database fields of the table.
                
                - Include foreign keys and reference fields.

            - C{records} (optional)
                
                  - C{[{"Table1Field1": val1,...}]}
                  
                  - A list of dictionaries containing key, value pairs for fields
                    to be created or updated
                  
                  - Includes foreign key and reference field values.
                  
                  - Contains the C{id} of the record to be updated.
                  
                  - The keys are encoded as C{Table1Field1}.
              
                  - The C{records} can also be a list containing the C{id} of the
                    records to be deleted. 

        
        @type keywords: dict
        @param keywords: list of keys which have to be in C{arg}
        
        @raise DbSvcException: when a keyword is missing or when the table
        does not exist in the database.
        
        """            
        # check that the request is well formed
        for kword in keywords:
            if kword not in arg:
                raise DbSvcException(KEYWORD_MISSING % kword)
        
        table = arg['tableName']
        self._is_table_in_db(table)


    def _encode_query(self, li):
        """Encode the query send by the client to a web2py C{Query}.
        
        @type li: list
        @param li: the query send by the client as a list of string.
        The following string are understood by the method::

            "[table1.field1] == [table2.field2]" or "[table1, field1] == [table2, field2]" 
            "[table1.field1] > n"
            "[table1.field1] like 'm%'"
            "[table1.field1] contains 'm%'"
            "[table1.field1] startswith 'm%'"
            "[table1.field1].lower() like 'm%'"
            "[table1.field1] belongs (1, 2, 3)"
            "([table1.field1] == [table2.field2]) | ([table1.field2] == [table2.field1])"

        @rtype: tuple of gluon.dal.Query
        @return:
            - the first element is the full query with all conditions or an empty string.
            - the second element contains only the left join conditions or an empty string.
            - the third element contains only the filter conditions or an empty string.
        
        @note: All elements of the list are ANDED in the web2py query
        
        @note: comparison operators are C{==, !=, <, >, <=, >=}
        
        @note: Boolean operators are C{&, |}

        """
        query, join, filter = "", "", ""
        
        if not li:
            return query
        
        for el in li:
            
            # transform bracket "[table1.field1]" as "db.table1.field1"
            s = REG_BRACKET.sub(r'db.\1.\2', el)
            
            # transform contains 'm%' as .contains("m%')
            # protection against EOL, TAB, ...
            for op in ['belongs', 'contains', 'like', 'startswith']:
                if op in s:
                    s = s.replace('\n', '').replace('\t', '')
                    s = re.sub(r" +(%s) +(.+)" %op, r'.\1(\2)', s)
            
            # separate join from filter queries
            if REG_IS_JOIN.match(s):
                join = ("(%s) & (%s)" % (join, s) if join else s)
                
            else:
                filter = ("(%s) & (%s)" % (filter, s) if filter else s)
            
            # build the query
            query = ("(%s) & (%s)" % (query, s) if query else s)

        if query:
            query = eval(query, {}, {"db": current.globalenv['db']})
        
        if join:
            join = eval(join, {}, {"db": current.globalenv['db']})
        
        if filter:
            filter = eval(filter, {}, {"db": current.globalenv['db']})
        
        return (query, join, filter)    

    
    def _get_record(self, table, id):
        """Get the record C{id} located in the database C{table}.
        
        @type table: gluon.dal.Table
        @param table:
        
        @type id: int
        @param id:
        
        @rtype: dict
        @return: key, value pairs where each key corresponds
        to a field of the table. Foreign keys are resolved.
        The key are encoded as C{Table1Field1}.
        
        """
        db = current.globalenv['db']
        record = {}

        for field in db[table].fields:
            key = encode_field(table, field)
            record[key] = db[table][id][field]

            if is_foreign_field(db[table][field]):
                k_table, k_field, k_query = get_foreign_field(db[table][field])
                
                k_key = encode_field(k_table, k_field)
                k_id = db[table][id][field]
                record[k_key] = db[k_table][k_id][k_field]
                
        return record


    def _is_field_in_table(self, table, field):
        """Check that the C{field} belongs to the C{table}.

        @type table: gluon.dal.Table
        @param table:
        
        @type field: gluon.dal.Field
        @param field:
        
        @rtype: bool
        
        """
        return field in current.globalenv['db'][table].fields
    
    
    def _is_fields_values_valid(self, table, fields):
        """Check each C{field} value against its C{validators}.
        

        @type table: gluon.dal.Table
        @param table:
        
        @type fields: dict
        @param fields: key, value pairs where each key corresponds
        to a C{gluon.dal.Field}
        
        @rtype: dict
        @return: key, value pairs for each field with errors.
        The key is encoded as C{Table1Field1} and the value is a string
        describing the error message.

        """
        di = {}
        db = current.globalenv['db']
        
        for (field, val) in fields.items():
        
            # NOTE: checks are not applied to foreign fields since 
            # consistency is garanty by the interface via the use of
            # combobox. This approach help when running with alias table
            # since validation crashed in that case.
            if is_foreign_field(db[table][field]):
                continue

            errmsg = db[table][field].validate(val)[1]
            if errmsg:
                di[encode_field(table, field)] = errmsg

        return di


    def _is_table_in_db(self, tablename):
        """Check that the table exists in the database.

        @type tablename: str
        @param tablename: name of the table
        
        @raise DbSvcException: when the table does not exists
        
        @note: The method works with regular and alias tables.
        
        """
        db = current.globalenv['db']
        tablenames =  [table._tablename for table in get_all_tables(db)]
        
        if tablename not in tablenames:
            raise DbSvcException(TABLE_NOT_IN_DB % table)


    def _prepare_records(self, arg):
        """Prepare the transaction data dictionary, C{arg}, for its insertion
        in the database (create or update).

        The method removes fields in the record which do not belong to the
        table since the JsonStore send the foreign key and the reference field.
        The latter are not used in create / update transaction.
        
        The method also validates each value.

        @type arg: dict
        @param arg: transaction data

            - C{tableName}
              the name of the table in the database
                
            - C{dbFields}
                
                - C{[["table1", "field1"], ... ]}
                
                - The list of database fields of the table.
                
                - Include foreign keys and reference fields.

            - C{records}
                
                  - C{[{"Table1Field1": val1,...}]}
                  
                  - A list of dictionaries containing key, value pairs for fields
                    to be created or updated
                  
                  - Includes foreign key and reference field values.
                  
                  - Contains the C{id} of the record to be updated.
                  
                  - The keys are encoded as C{Table1Field1}.
              
        @rtype: gluon.storage.Storage
        @return:
         
            - C{Storage(errors=[], records=[])}
        
            - One to one correspondence between the errors 
              and the records lists
            
            - C{records} is a list of dictionaries containing key value pairs
              where the key corresponds to a database field. 
              The keys are encoded as C{Table1Field1}.
              
            - The C{error} is C{None} when the C{record} is validated.
              Otherwise the error is a dictionary 
              C{{Table1Field1: "error message", ...}}
              
        """
        data = Storage(errors=[], records=[])
        table = arg['tableName']

        for record in arg[ROOT]:        
            fields = {}
            
            # Remove fields in the record which do not belong to the table.
            # The JsonStore send the foreign key and the the pointing field
            # when manipulating table with foreign keys. the latter belongs
            # to another table.
            for table_field in record:
                tablename, fieldname = decode_field(table_field)
                if tablename == table:
                    fields[fieldname.encode('utf8')] = record[table_field]

            data.errors.append(None)
            data.records.append(fields)
            
            # Validate field contents
            di = self._is_fields_values_valid(table, fields)
            if di:
                table_id = encode_field(table, 'id')
                record[table_id] = ""
                data.records[-1] = record
                data.errors[-1] = di

        return data

    
    def create(self, arg):
        """Create new records defined by the transaction data C{arg}.
        Several transactions of the same type are processed together.
        
        @type arg: dict
        @param arg: transaction data

            - C{tableName}
              the name of the tablename in the database
                
            - C{dbFields}
                
                - C{[["table1", "field1"], ... ]}
                
                - The list of database fields of the tablename.
                
                - Include foreign keys and reference fields.

            - C{records}
                
                  - C{[{"Table1Field1": val1,...}]}
                  
                  - A list of dictionaries containing key, value pairs for fields
                    to be created.
                  
                  - Includes foreign key and reference field values.
                  
                  - The keys are encoded as C{Table1Field1}.
              
        @rtype: dict
        @return:
            
            - C{{success: True, records:[{Table1Field1: val, ... }, ... ], 
              msg: 'blalbla'}}
            
            - C{records} is a list of dictionary containing the complete
              key, value pairs for the C{record} created in the database.
            
            - The record keys are encoded as C{Table1Field1}.
            
            - C{{success: False, records:[{Table1Field1: val, ... }, ... ],
              errors: 'blabla', }}
              when at least a field value is not validated or when the transaction
              is kill by the callback C{table._before_insert}.  
              The full transaction is aborted. 
              The errors contains the error message return by callback or
              a dictionary returns by the validator. The key is the field 
              identifier encoded as C{Table1Field1} while the value is
              the error message.
            
        """
        self.dbg("Start DbSvc.create")

        self._check_request(arg)
        data = self._prepare_records(arg)

        # Abort the full transaction if at least one record is in error
        for error in data.errors:
            if error:
                i = data.errors.index(error)
                di = encode_validator_errors(error)
                return {SUCCESS: False, "errors": di, ROOT: data.records[i]}
        
        # insert new records  
        ids, tablename, records = [], arg['tableName'], []
        table = current.globalenv['db'][tablename]
        
        for fields in data.records:
            
            id = table.insert (**fields)

            # operation can be reject by callbacks table._before_insert
            if not id:
                msg = REJECT_BY_CALLBACK
                if CALLBACK_ERRORS in table:
                    msg = table._callback_errors
                    
                return {SUCCESS: False, "errors": msg, ROOT: fields}
            
            # return the complete record
            # mandatory to display the new record in the grid.
            record = self._get_record(tablename, id)
            
            ids.append(str(id))
            records.append(record)
        
        txt = ', '.join(ids)
        self.dbg("End DbSvc.create.", RECORDS_INSERTED % txt)
        return {SUCCESS: True, 'msg': RECORDS_INSERTED % txt, ROOT: records}

    
    def destroy(self, arg):
        """Delete existing records defined by the transaction data C{arg}.
        Several transactions of the same type are processed together.
        
        @type arg: dict
        @param arg: transaction data

            - C{tableName}
              the name of the tablename in the database
                
            - C{dbFields}
                
                - C{[["table1", "field1"], ... ]}
                
                - The list of database fields of the tablename.
                
                - Include foreign keys and reference fields.

            - C{records} the list of C{id} for the records
              to be deleted, C{[id1, id2, ... ]}. 

        @rtype: dict
        @return:
            
            - C{{success: True, records:[{Table1Id1: id}, ...], msg: 'blalbla'}}
            
            - C{records} is a list of dictionary containing the C{id}
              of the delete records: C{[{TableId:xx}, ..]}.
            
            - The record key are encoded as C{Table1Field1}.
            
            - C{{success: False, records:[{Table1Field1: id}, ... ],
              errors: 'blabla', }}
              when at least one record does not exist or when the delete 
              operation is rejected by the callback C{table._before_delete}.
              The full transaction is aborted.
              The errors contains the error messages return by the callback.

        """
        self.dbg("Start DbSvc.destroy")

        self._check_request(arg)
        
        ids, tablename, records = [], arg['tableName'], []
        table_id = encode_field(tablename, 'id')
        db = current.globalenv['db']
        table = db[tablename]
        
        # Abort the transaction is at least one record does not exists
        for id in arg[ROOT]:
            if not table[id]:
                txt = RECORD_NOT_IN_DB % (id, tablename)
                return {"success": False, "msg": txt, ROOT: {table_id: id}}

        # delete records        
        for id in arg[ROOT]:
            
            rep = db(table.id==id).delete()
            
            # operation can be reject by callbacks table._before_delete
            if not rep:
                msg = REJECT_BY_CALLBACK
                if CALLBACK_ERRORS in table:
                    msg = table._callback_errors
                return {SUCCESS: False, "errors": msg, ROOT: {table_id: id}}

            ids.append(str(id))
            records.append({table_id: id})
            
        txt = ', '.join(ids)
        self.dbg("End DbSvc.destroy")
        return {"success": True, "msg": RECORDS_DELETED % txt,  ROOT: records}
    
    
    def read(self, arg):
        """Read the content of a table as specified in the transaction data C{arg}.
        The C{arg} dictionary contains the following keys:

        @type arg: dict
        @param arg: transaction data

            - C{tableName}
              the name of the table in the database
                
            - C{dbFields}
                
                - C{[["table1", "field1"], ... ]}
                
                - The list of database fields of the table.
                
                - Include foreign keys and reference fields.
                            
            - C{where} (optional) 
                
                - C{['db.table1.field1  == db.table2.id', ... ]}
                
                - a list of string to build the inner join 
                  when handling foreign key.
              
                - The string syntax follows the web2py convention:
                  C{db.table1.field1  == db.table2.id}.
                   
                - An AND is performed between the different 
                  elements of the list.

            - C{orderby} (optional)
            
                - C{[("table1", "field1", "dir"), ("table2", "field3", "dir"), ...]}
                
                - a list of tuples to build the order directive.
                
                - the third argument is equal to C{"ASC"} or C{"DESC"}.
                
        The transaction data can also contains parameters useful for paging.
        For more detail see  Ext.PagingToolbar:
            - start
            - limit
            - sort
            - dir

        @rtype: dict
        @return:

            - C{{success: True, records: [{TableField: value, ...}, ...]}}
            
            - C{records} is a list of dictionary containing the key value
              pairs for the record found in the database.
              
            - the record keys are encoded as C{Table1Field1}.
            
        """
        self.dbg("Start DbSvc.read")
        
        self._check_request(arg, ('tableName', 'dbFields'))
        db = current.globalenv['db']

        # build the list of SQLField object
        fields = []
        for table, field in arg['dbFields']:
            if not self._is_field_in_table(table, field):
                raise DbSvcException(FIELD_NOT_IN_DB % (table, field))
            fields.append(db[table][field])

        # build the inner join and filter conditions
        query, q_join, q_filter = '', '', ''
        if 'where' in arg:
            query, q_join, q_filter = self._encode_query(arg['where'])

        # Count the total numbers of records in the table
        # or in the set when filter conditions are applied
        tablename = arg['tableName']
        if not q_filter:
            nrecords = db(db[tablename]).count()
        
        elif q_filter and tablename in db:
            nrecords = db(q_filter).count()
            
        # NOTE: the usual method db(db.table).count() failed with alias table
        # it is why we use a more complicated approach.
        elif q_filter and tablename not in db:
            nrecords = len(db(q_filter).select(db[tablename].id))
            
        # handle paging options (see Ext.PagingToolbar)
        kwargs = {}
        if 'start' in arg and 'limit' in arg:
            start = int(arg['start'])
            limit = int(arg['limit'])
            kwargs['limitby'] = (start, start+limit)

        # handle the orderby directive ['[table1, field2, dir]', ....]
        if 'orderby' in arg:
            kwargs['orderby'] = [] 
            for tablename, fieldname, dir in arg['orderby']:
                el = db[tablename][fieldname]
                if dir == 'DESC':
                    el = ~el
                kwargs['orderby'].append(el)
                     
        # interrogate the database and serialize the records as a list:
        # [{TableField: value, }, {...}, ...]
        rows = db(query).select(*fields, **kwargs)
        li = rows_serializer(rows)
               
        self.dbg('End DbSvc.read', li)
        return {SUCCESS: True, TOTAL: nrecords, ROOT: li}


    def update(self, arg):
        """Update records defined by the transaction data C{arg}.
        Several transactions of the same type are processed together.
        
        @type arg: dict
        @param arg: transaction data

            - C{tableName}
              the name of the tablename in the database
                
            - C{dbFields}
                
                - C{[["table1", "field1"], ... ]}
                
                - The list of database fields of the tablename.
                
                - Include foreign keys and reference fields.

            - C{records}
                
                  - C{[{"Table1Field1": val1,...}]}
                  
                  - A list of dictionaries containing key, value pairs for fields
                    to be updated
                  
                  - Includes foreign key and reference field values.
                  
                  - Contains the C{id} of the record to be updated.
                  
                  - The keys are encoded as C{Table1Field1}.
              
        @rtype: dict
        @return:
            
            - C{{success: True, records:[{Table1Field1: val, ... }, ... ], 
              msg: 'blalbla'}}
            
            - C{records} is a list of dictionary containing the complete
              key, value pairs for the C{record} updated in the database.
            
            - The record keys are encoded as C{Table1Field1}.

            - C{{success: False, records:[{Table1Field1: val, ... }, ... ],
              errors: 'blabla', }}
              when at least a field value is not validated or when the transaction
              is kill by the callback C{table._before_update}.  
              The full transaction is aborted. 
              The errors contains the error message return by callback or
              a dictionary returns by the validator. The key is the field 
              identifier encoded as C{Table1Field1} while the value is
              the error message.
            
        """
        self.dbg("Start DbSvc.update.")

        self._check_request(arg)
        data = self._prepare_records(arg)

        # Abort the full transaction if at least one record is in error
        for error in data.errors:
            if error:
                i = data.errors.index(error)
                di = encode_validator_errors(error)
                return {SUCCESS: False, "errors": di, ROOT: data.records[i]}

        # update records  
        ids, tablename, records = [], arg['tableName'], []
        db = current.globalenv['db']
        table = db[tablename]
        
        for fields in data.records:

            # the client send modified fields only
            # update the database accordingly        
            id = fields['id']
            rep = db(table.id==id).update(**fields)

            # operation can be reject by callbacks table._before_update
            if not rep:
                msg = REJECT_BY_CALLBACK
                if CALLBACK_ERRORS in table:
                    msg = table._callback_errors
                    return {SUCCESS: False, "errors": msg, ROOT: fields}

            # return the complete record
            # mandatory to display the new record in the grid.
            record = self._get_record(tablename, id) 

            ids.append(str(id))
            records.append(record)
            
        txt = ', '.join(ids)
        self.dbg("End DbSvc.update.")
        return {SUCCESS: True, "msg": RECORDS_UPDATED % txt, ROOT: records}
