""" configurators for the Ext JS components. 

@author: R. Le Gac

"""

from gluon import current
from gluon.tools import PluginManager

MSG_XTYPE = 'Invalid keyword "xtype"'


class ExtJSException(BaseException): pass


class Base(dict):
    """Base class for Ext JS configurator.
    
    The base class comes with two methods C{append_items} 
    and C{append_plugins}. They are required to append C{items}
    and C{plugins}.
    
    @note: A protection is set to avoid changing the C{xtype} of the element.

    """
    xtype = None
    
    def __init__(self, **kwargs):
        """
        
        @keyword kwargs: any Ext JS configuration parameter of the target class
     
        @raise ExtJSException: if the keyword C{xtype} 
        is in the keyword arguments.
        
        """
        if 'xtype' in kwargs:
            raise ExtJSException(MSG_XTYPE)

        self['xtype'] = self.xtype

        items = kwargs.pop('items', None)
        if items:
            self._append('items', items)
        
        plugins = kwargs.pop('plugins', None)
        if plugins:
            self._append('plugins', plugins)
            
        self.update(kwargs)


    def _append(self, key, t):
        li = list(t)
        if key in self:
            self[key].extend(li)
        else:
            self[key] = li


    def append_items(self, *args):
        self._append('items', args)

        
    def append_plugins(self, *args):
        self._append('plugins', args)


class ArrayStore(Base):
    """Configurator for the C{Ext.data.ArrayStore}."""
    xtype='arraystore'


class CheckBox(Base):
    """Configurator for C{Ext.form.Checkbox}."""
    xtype = 'checkbox'


class ComboBox(Base):
    """Configurator for C{App.form.CombBox}."""
    xtype = 'xcombobox'


class CompositeField(Base):
    """Configurator for C{Ext.form.CompositeField}."""
    xtype = 'compositefield'


class DateField(Base):
    """Configurator for C{Ext.form.DateField}."""
    xtype = 'datefield'


class DirectStore(Base):
    """Configurator for C{App.data.DirectStore}."""
    xtype = 'xdirectstore'


class Field(Base):
    """Configurator for C{Ext.form.Field}."""
    xtype = 'field'


class FieldSet(Base):
    """Configurator for C{Ext.form.FieldSet}.

    @note: By default, the layout for children Field is set to C{anchor 99%}.
    It can be overwritten using the keywords defaults.
       
    """
    xtype = 'fieldset'
    
    def __init__(self, **kwargs):
        self['defaults'] = {'anchor': '99%'}
        Base.__init__(self, **kwargs)


class FormPanel(Base):
    """Configurator for C{App.form.FormPanel}."""
    xtype = 'xform'


class GridColumn(dict):
    """Configurator for C{Ext.grid.Column}."""


class GridColumnModel(list):
    """Configurator for C{Ext.grid.ColumnModel}.
    
    @note:It is  implemented as a C{list} of L{GridColumn}.
    
    """


class GridFilter(Base):
    """Configurator for C{App.grid.GridFilter}."""
    xtype = 'xgridfilter'


class GridPanel(Base):
    """Configurator for C{App.grid.Grid}."""
    xtype = 'xgrid'


class GridRowNumberer(Base):
    """Configurator for C{Ext.grid.RowNumberer}."""
    xtype = 'rownumberer'


class GridTemplateColumn(Base):
    """Configurator for C{Ext.grid.TemplateColumn}."""
    xtype = 'templatecolumn'


class GridWithFilter(Base):
    """Configurator for C{App.grid.GridWithFilter}."""
    xtype = 'xgridwithfilter'


class JsonStore(Base):
    """Configurator for the C{Ext.data.JsonStore}."""
    xtype='jsonstore'


class NumberField(Base):
    """Configurator for C{Ext.form.NumberField}."""
    xtype = 'numberfield'


class Panel(Base):
    """Configurator for C{Ext.Panel}."""
    xtype = 'panel'


class PanelWithUrlSelector(Base):
    """Configurator for C{App.PanelWithUrlSelector}."""
    xtype = 'xpanelwithurlselector'


class Spacer(Base):
    """Configurator for C{Ext.Spacer}."""
    xtype = 'spacer'


class Store(Base):
    """Configurator for the C{Ext.data.Store}."""
    xtype='store'


class TabPanel(Base):
    """Configurator for C{Ext.TabPanel}."""
    xtype = 'tabpanel'


class TextArea(Base):
    """Configurator for C{Ext.form.TextArea}.
    
    @note: Regular expression for the validation can be written as a string.
    
    """
    xtype = 'textarea'

    def __init__(self, **kwargs):
        self['plugins'] = ['pRegExp']
        Base.__init__(self, **kwargs)

    
class TextField(Base):
    """Configurator for C{Ext.form.TextField}.
    
    @note: Regular expression for the validation can be written as a string.
        
    """
    xtype = 'textfield'

    def __init__(self, **kwargs):
        self['plugins'] = ['pRegExp']
        Base.__init__(self, **kwargs)
        
    
class TimeField(Base):
    """Configurator for C{Ext.form.TimeField}."""
    xtype = 'timefield'


class Viewport(Base):
    """Configurator for C{App.Viewport}."""
    xtype = 'xviewport'


class XmlStore(Base):
    """Configurator for the C{Ext.data.XmlStore}."""
    xtype='xmlstore'
    