""" the fields modifier

@author: R. Le Gac 

"""

from gluon.storage import Storage
from modifier import Modifier


MODIFIER_FIELDS = 'modifier_fields'


class FieldsModifier(Modifier):
    """Modifier to customise the field widgets associated to a database table.
    
    """
    
    def __init__(self, tablename):
        """Initialize the fields modifier associated to the database
        table C{tablename}.
        
        The data structure of the fields modifier can be access 
        outside the class:
        
            >>> p = PluginManager('dbui')
            >>> modifier = p.dbui['modifier_fields'][tablename]

        or internally via the attribute C{data}.
        
        The keys of the C{data} structure are:
        
            - C{extjs_fields} (dict) Ext JS configuration options 
              for individual C{field}.
            
            - C{composite_fields.main} (list) the first field 
              of the CompositeField.
            
            - C{composite_fields.others} (list of list) the others field of
              a CompositeField.
            
            - There is one to one correspondence between the C{composite_fields.main}
              and the C{composite_fields.others} lists.

        @type tablename: str
        @param tablename: name of the database table
        
        """
        Modifier.__init__(self, MODIFIER_FIELDS, tablename)

        if 'extjs_fields' not in self.data:
            self.data.extjs_fields = {}
        
            self.data.composite_fields = Storage()
            self.data.composite_fields.main = []
            self.data.composite_fields.others = []


    def append_plugins(self, *plugins):
        """Overwrite the method of the base class since it make
        no sense for this modifier.
        
        """


    def configure(self, **extjs):
        """Overwrite the method of the base class since it make
        no sense for this modifier.
        
        """
        
        
    def configure_field(self, field, **kwargs):
        """Update the configuration options of the Ext JS widget 
        associated to the C{field}.        
        Existing value of the configuration options are replace by those 
        defined by the keyword arguments.
        
        @type field: str
        @param field: name of the database field. 
        It should belong to the database table defined in the constructor.
        
        """
        self.data.extjs_fields[field] = kwargs


    def merge_fields(self, *fields, **kwargs):
        """Merge fields in a CompositeField. 
        
        @note:
            - Each C{field} is identified by its database field name.
            - C{field} should belong to the database table defined in the constructor.
            - The first C{field} is the main field of the C{CompositeField}.
            - The keyword arguments contains the configuration options of
              the C{Ext.form.CompositeField} widget.
                
        """

        # define the main field as the first field in the list
        main_field = fields[0]
        other_fields = fields[1:]
          
        # In order to simplify the processing keep a list of
        # the main field and a list of the other fields.
        self.data.composite_fields.main.append(main_field)
        self.data.composite_fields.others.extend(other_fields)
        
        di = Storage(fields=fields, extjs=kwargs)
        self.data.composite_fields[main_field] = di
