""" the form modifier

@author: R. Le Gac

"""
import copy

from fieldsmodifier import FieldsModifier
from gluon.storage import Storage
from helper import (as_list, 
                    encode_field, 
                    get_foreign_field, 
                    get_store_id,
                    is_foreign_field)
from modifier import Modifier


MODIFIER_FORMS = 'modifier_forms'


def configure_forms(db, **extjs):
    """Apply to all form widgets the same configuration options. 
    Useful to set plugins in one go, for example.
    
    @type db: gluon.dal.DAL
    @param db:
    
    @note: The keyword arguments define values for the Ext JS configuration 
    options of he C{Ext.form.FormPanel} class.
    
    """
    for tablename in db.tables:
        fm = FormModifier(tablename)
        di = copy.deepcopy(extjs)
        fm.configure(**di)



class FormModifier(Modifier):
    """Modifier to customise the form widget associated to a database table.
    
    """
    
    def __init__(self, tablename):
        """Initialize the form modifier associated to the database
        table C{tablename}.
        
        The data structure of the form modifier can be access 
        outside the class:
        
            >>> p = PluginManager('dbui')
            >>> modifier = p.dbui['modifier_forms'][tablename]

        or internally via the attribute C{data}.
        
        The keys of the C{data} structure are:
        
            - C{extjs} (dict) Ext JS configuration options for the form widget.
            
            - C{field_sets} (list of Storage) each C{Storage} defined a 
              C{FieldSets}. It has two keys C{fields} and C{extjs}. 
              The former is a list of field names belonging to the table
              defined in the constructor. The latter is a dictionary with
              the configuration options of the C{Ext.form.FieldSet} widget.
            
            - C{hidden_fields} (list)  list of field names to be hidden in 
            the form. The fields have to belong to the table defined in
            the constructor.
            
            - C{mapper} (None or function) reference to the function used
            to organize the C{FieldSet} in C{Tabs}, for example.

        @type tablename: str
        @param tablename: name of the database table
        
        """
        Modifier.__init__(self, MODIFIER_FORMS, tablename)

        if 'field_sets' not in self.data:
            self.data.field_sets = []
            self.data.hidden_fields = []
            self.data.mapper = None


    def link_comboboxes(self, **kwargs):
        """Link two fields in such a way that the values available for the 
        second one depend on the value selected for the first one.
        
            - This first field is the master while the second one is the slave.
            - Often the master and slave fields are reference fields.
            - An external table contains the data defining the relation between the
              master and the slave.
            - The external table must have the master and slave fields
            - Several slaves can be attached to the same master::
                
                    fieldsModifier = fieldsModifier('mytable')
                    fieldsModifier.link_comboxes(master=virtdb.mytable.myfield1,
                                                 slave=virtdb.mytable.myfield2,
                                                 masterHasSlaveData='anotherTable')
                    fieldsModifier.link_comboxes(master=virtdb.mytable.myfield1,
                                                 slave=virtdb.mytable.myfield3,
                                                 masterHasSlaveData='anotherTable')
                                                     
            - This method modify deeply the field configuration transforming
              the Ext.form.comboBox into App.form.LinkedComboBox.

        @type master: gluon.dal.Field
        @keyword master: the field used as master.
        It has to be defined in the table referenced in the constructor
        and in the C{masterHasSlaveData} table.

        @type slave: gluon.dal.Field
        @keyword slave: the field used as salve.
        It has to be defined in the table referenced in the constructor
        and in the C{masterHasSlaveData} table.

        @type masterHasSlaveData: str
        @keyword masterHasSlaveData: name of the external database table.
        Its contains the data defining the relation between master and slave. 
        It should contains fields with the name of the master and slave ones.
                        
        @type masterExtJS: dict
        @keyword masterExtJS: configuration options dictionary for the 
        Ext.form.Combobox associated to the master reference field. 
        Might be useful to set the master in readOnly mode for example.

        @type slaveExtJS: dict
        @keyword slaveExtJS: configuration option dictionary for the
        Ext.form.Combobox associated to the slave reference field. 
        Might be useful to set the salve in readOnly mode for example.
                
        """
        # short cuts
        myTable = kwargs['master'].tablename
        exTable = kwargs['masterHasSlaveData']

        fieldsModifier = FieldsModifier(myTable)
        
        # Configure the master field
        # Skip this step if the master is already defined -- multi slaves
        master = kwargs['master']
        data = fieldsModifier.data.extjs_fields
        
        is_master = master in data and \
                    'xtype' in data[master] and \
                    data[master]['xtype'] == 'xlinkedcombobox'
                     
        if not is_master: 
                   
            extjs = {}
            if 'masterExtJS' in kwargs:
                extjs = kwargs['masterExtJS']
            
            if is_foreign_field(master):
                kTable, kField, kId = get_foreign_field(master)
            
            else:
                kTable = exTable
                kField = master.name
            
            fieldsModifier.configure_field(master.name,
                                           displayField=encode_field(kTable, kField),
                                           valueField=encode_field(exTable, master.name),
                                           role='master',
                                           store=get_store_id(exTable),
                                           xtype='xlinkedcombobox',
                                           **extjs)

        # Configure the slave field
        extjs = {}
        if 'slaveExtJS' in kwargs:
            extjs = kwargs['slaveExtJS']

        slave = kwargs['slave']
        if is_foreign_field(slave):
            kTable, kField, kId = get_foreign_field(slave)
        
        else:
            kTable = exTable
            kField = slave.name
        
        fieldsModifier.configure_field(slave.name,
                                       displayField=encode_field(kTable, kField),
                                       valueField=encode_field(exTable, slave.name),
                                       mode='local',
                                       role='slave',
                                       store=get_store_id(exTable),
                                       xtype='xlinkedcombobox',
                                       **extjs)
        

    def merge_fields(self, *fields, **kwargs):
        """Merge fields in a L{FieldSet} corresponding to 
        an C{Ext.form.FieldSet} widget. 
        
        Each C{field} is either a string identifying the database field by its name
        or an instance of the Ext JS configurator like the C{Spacer}.
        
        The keyword argument define values for the configuration options of
        the C{Ext.form.FieldSet} widget.
        
        """
        # by default the field occupies all the field set space
        extjs = {'defaults': {'anchor': '100%'}}
        extjs.update(kwargs)

        di = Storage(fields=fields, extjs= extjs)
        self.data.field_sets.append(di)


    def hide_fields(self, *fields):
        """Hide the list of fields in the form.
        Each field is identified by its database field name.
        
        """
        for el in fields:
            if el not in self.data.hidden_fields:
                self.data.hidden_fields.append(el)


    def set_mapper(self, func):
        """Link the mapper tool to the function func.
        
        A mapper tool is related to the organization of the field/fieldest
        in the form. By default the form layout is a column where fields are 
        pushed from the top to the bottom. By using the layout configuration 
        option it is possible to organize the field/fieldset in horizontal
        box, vertical box, columns or in table.
        
        This approach does not allow to map field/fieldset in panels
        contains in the form, like an Ext.TabPanel. The mapper tool allows to 
        handle this case. It is a function with one arguments, the list of 
        field/fieldset. The function return a dictionary with the structure of
        panels embedded in the form an the fields they contained.

        Example of mapper can be found in the L{mapper} module.
        
        """
        self.data.mapper = func