"""the grid modifier.

@author: R. Le Gac

"""
import copy

from gluon.storage import Storage
from modifier import Modifier


MODIFIER_GRIDS = 'modifier_grids'

MSG_HEADER_MISSING = 'configuration option header is missing in extjs'
MSG_INVALID_KEYWORD = 'invalid keyword %s'
MSG_POSITION_MISSING = 'keyword argument position is missing'
MSG_TPL_MISSING = 'configuration option tpl is missing in extjs'


def configure_grids(db, **extjs):
    """Apply to all grid widgets the same configuration options. 
    Useful to set plugins in one go, for example.
    
    @type db: gluon.dal.DAL
    @param db:
    
    The keyword arguments contain the configuration options for
    the Ext.grid.GridPanel.
    
    """
    for tablename in db.tables:
        gm = GridModifier(tablename)
        di = copy.deepcopy(extjs)
        gm.configure(**di)


class GridModifierException(BaseException): pass

    
class GridModifier(Modifier):
    """Modifier to customize the grid widget associated to a table.
    
    """
    
    def __init__(self, tablename):
        """Initialize the grid modifier associated to the database
        table C{tablename}.
        
        The data structure of the grid modifier can be access 
        outside the class:
        
            >>> p = PluginManager('dbui')
            >>> modifier = p.dbui['modifier_grids'][tablename]

        or internally via the attribute C{data}.
        
        The keys of the C{data} structure are:
        
            - C{extjs} (dict) Ext JS configuration options for the grid widget.
            
            - C{configure_columns} (dict) the Ext JS configuration options for 
              the C{Ext.grid.Column} widget. The column is identify by its 
              C{fieldname}.
            
            - C{delete_columns} (list)
            
            - C{grid_filters} (storage)
            
            - C{hidden_columns} (list)
            
            - C{row_numbering} (bool)
            
            - C{template_columns} (list

        @type tablename: str
        @param tablename: name of the database table
        
        """
        Modifier.__init__(self, MODIFIER_GRIDS, tablename)
        
        if 'configure_columns' not in self.data:
            self.data.configure_columns = {}
            self.data.delete_columns = []
            self.data.grid_filters = Storage(filters=[],
                                             extjs_filters=[], 
                                             extjs={})
            self.data.hidden_columns = []
            self.data.row_numbering = False
            self.data.template_columns = []

                   
    def add_column(self, ktable, kfield, position=0, extjs={}):
        """NOT YET IMPLEMENTED
        
        Add a column displaying the database field kfield of the 
        foreign table ktable. 
        
        The keyword argument position is a number defining where the
        column will be insert. The first column is at 0.
        
        The keyword argument ExtJS is a dictionary containing
        configuration options of the underlying Ext.grid.Column.
        
        """


    def configure_column(self, fieldname, **kwargs):
        """Modify the configuration options for a given column.
        Each column is identified by its database field name.
        
        This method is useful to fine tune column width for example.

        @type fieldname: str
        @param fieldname: name of a database field. 
        It has to belong to the table defined in the constructor.
        
        The keyword arguments defines the configuration options
        for the associated C{Ext.grid.Column} widget.

        """
        self.data.configure_columns[fieldname] = kwargs

        
    def delete_columns(self, *fields):
        """Delete the columns in the C{Ext.grid.GridPanel} widget.
        Each column is identified by its database field name.
        
        """
        for el in fields:
            if el not in self.data.delete_columns:
                self.data.delete_columns.append(el)

        
    def hide_columns(self, *fields):
        """Hide the columns in the C{Ext.grid.GridPanel} widget.
        Each column is identified by its database field name.
        
        """
        for el in fields:
            if el not in self.data.hidden_columns:
                self.data.hidden_columns.append(el)
        
        
    def merge_columns(self, *fields, **kwargs):
        """Merge several fields in a single C{TemplateColumn}.
        The TemplateColumn is rendered by the C{Ext.grid.TemplateColumn}.
        Each field is identified by its database field name.
        
        @keyword header: (str) the header of the TemplateColumn
        
        @keyword tpl: (list) the template defining how fields are displayed
        in the TemplateColumn.
            - The syntax of the template is is defined in the documentation
              of C{Ext.XTemplate} constructor.        
            - There is some limitation when defining the template in python,
              namely the use of the template member function.
            - Database fields are identified using the C{CapWord} syntax: 
              C{TablenameFieldname}. It works for standard and reference fields.
            - HTML balises are understood.
            - Example of a template list::
        
                tpl = ['<b>{PeopleLast_name}</b><br>',
                       '{PeopleFirst_name}<br>',
                       '<tpl if={PeoplePhone_number}>{PeoplePhone_number}</tpl>']
        
        @keyword autohide: (bool) hide the columns associated to the fields
        used in the TemplateColumn when set to C{True}.

        @keyword position: (int) is a number defining where the
        column will be inserted in the grid widget. 
        The first column is at 0.
        
        @keyword extjs: (dict) contains the Ext JS configuration options
        of the C{Ext.grid.TemplateColumn}.
        
        """
        keywords = ['autohide', 'position']
        extjs = dict(kwargs)
        
        for key in keywords:
            if key in extjs:
                del extjs[key]
        
        if 'position' not in kwargs:
            raise GridModifierException(MSG_POSITION_MISSING)
        
        if 'header' not in kwargs:
            raise GridModifierException(MSG_HEADER_MISSING)

        if 'tpl' not in kwargs:
            raise GridModifierException(MSG_TPL_MISSING)
            
        di = Storage(extjs=extjs, position=kwargs['position'])
        self.data.template_columns.append(di)
        
        if 'autohide' in kwargs and kwargs['autohide']:
            self.hide_columns(*fields)


    def append_filter(self, filter, **kwargs):
        """Append a filter to the GridFilter.
        The filter will be add to the GridFilter which is a FieldSet
        located on the right side of the GridPanel. 
        A filter corresponds to a field in the FieldSet.
        
        @type filter: tuple
        @param filter: A filter is defined by a tuple containing three strings:
            - The database field name. The field has to
              belong to the database table defined in the constructor. 
              However, the more elaborate syntax, C{'table2.field1'}, 
              allows to use foreign field.
            - The operator, C{'=='}, C{'contains'}, ...
              Valid operator are defined in the method L{dbsvc.DbSvc._encode_query}
            - The comment for the tool tip.
            
        The keyword argument contains Ext JS configuration options
        for the associated Ext.form.Field.
        
        """
        self.data.grid_filters.filters.append(filter)
        self.data.grid_filters.extjs_filters.append(kwargs)


    def set_filters(self, *filters, **kwargs):
        """Define the GridFilter associated to the grid.
        The GridFitler is an C{Ext.form.Fieldset} located on the right side 
        of the GridPanel. It contains field widgets associated to each filter: 
        field, comboBox, dataField,..
        
        The C{filter} is defined by a tuple containing three strings:
            - The database field name. The field has to
              belong to the database table defined in the constructor. 
              However, the more elaborate syntax, C{'table2.field1'}, 
              allows to use reference field.
            - The operator, C{'=='}, C{'contains'}, ...
              Valid operator are defined in the method 
              L{dbsvc.DbSvc._encode_query}.
            - The comment for the tool tip.
        
        The keyword argument contains the Ext JS configuration options for the 
        C{Ext.form.FieldSet}.

        Filter can be defined and tuned individually 
        using the method L{append_filter}.
        
        """
        for el in filters:
            self.append_filter(el)
            
        self.data.grid_filters.extjs=kwargs


    def set_rownumbering(self):
        """Activate the row numbering.
        Row number is displayed in the first column of the grid.
        
        """
        self.data.row_numbering = True
