""" a series of helper functions

@author: R. Le Gac

"""

import os
import re
import subprocess
import tempfile

from gluon import current
from gluon.dal import Row, Table
from gluon.http import HTTP
from gluon.validators import (IS_DATE_IN_RANGE,
                              IS_DATETIME_IN_RANGE,
                              IS_DECIMAL_IN_RANGE,
                              IS_EMPTY_OR,
                              IS_FLOAT_IN_RANGE,
                              IS_IN_DB,
                              IS_IN_SET,
                              IS_INT_IN_RANGE,
                              IS_MATCH,
                              IS_LENGTH)

GIT = '/usr/bin/git'
STOREID = '%sStore'


def as_list(val):
    """Encapsulated C{val} into a C{list} when C{val} is not an
    instance of a C{list}.
    
    @type val: any
    @param val: input value
    
    @rtype: list
    
    """
    if not isinstance(val, (list, tuple)):
         val = [val]
    return val


def decode_field(field):
    """Decode a string in CapitalizeWords into a list
    of words where all words are in lower case.
    The reverse function is L{encode_field}.
    
    @note: useful to decode database field name:
    C{"TableField"} or C{"TableFieldFunction"}
    
    @type field: str
    
    @rtype: tuple

    """ 
    li = []
    for c in field:
        if c.isupper():
            li.append(c.lower())
        elif li:
            li[-1] += c
    return tuple(li)


def dummy_row(table, value=''):
    """Create a dummy C{Row} for the C{table}. The row contains keys for 
    the table and for the reference tables. Each field are set to C{value}.

    The Row van be inserted in a gluon.dal.Rows::
        
        rows = db(query).select(orderby=db.table.field)
        empty_row = dummy_row(db.table)
        row.records.append(row)
    
    @type table: gluon.dal.Table
    @param table:
    
    @type value: string
    @param: value associated to each field either an empty string of None
    
    @rtype: gluon.dal.Row
    
    """
    db = table._db
    tablename = table._tablename
    
    row = Row()
    row[tablename] = Row()
    
    for field in table:
        row[tablename][field.name] = value
        
        if is_foreign_field(field):
            k_tablename, k_fieldname, k_id = get_foreign_field(field)
            if k_tablename not in row:
                for k, v in dummy_row(db[k_tablename], value=value).iteritems():
                    if k not in row:
                        row[k] = v
            
    return row

    
def encode_field(*args):
    """Encode the string arguments into a CapitalizeWords.
    The reverse function is L{decode_field}.

    @rtype: str
    
    @note: useful to build database field name like C{TableField} 
    or C{TableFieldFunc}.
    
    @note: The Ext JS ComboBox don't decode properly any separator between 
    table and field like table+field. Therefore choice for the encoding
    are very limited.
    
    """ 
    # NOTE: don't use the method title since foo_faa 
    # becomes Foo_Faa and you want Foo_faa.
    return ''.join([el[0].upper()+el[1:].lower() for el in args if len(el)])


def encode_validator_errors(error):
    """Encode and translate the validator errors.
    
    @type error: dict
    @param error: the key is the table field encode as C[TablenameFieldname]
    while the value is the error message return by the validator
    
    @rtype: dict
    @return: error message are translated and should be more friendly
    since it contains the field name with its error.
    
    """
    di = {}
    for k, v in error.iteritems():
        tablename, fieldname = decode_field(k)
        di[k] = "%s %s" % (current.T(fieldname.title()),v)
    return di


def get_all_tables(dal):
    """Get the list of names for all tables in the database, C{dal},
    including alias table.
    
    @type dal: gluon.dal.DAL
    @param dal:
    
    @rtype: list
    
    @note: It is recommend to use standard iterator provided by 
        the gluon.dal.DAL class. However by construction, they ignore 
        alias table.
    
    """
    li = [dal[attr] for attr in dir(dal) if isinstance(dal[attr], Table)]
    return li

    
def get_create_id(table, **kwargs):
    """Get the C{id} of the record identified by a set of C{field=value} 
    keyword arguments.
    If the record does not exist it is created with the values
    specified in the keywords arguments.

    Example::
    
        id = get_create_id(db.person, first_name="Joe", last_name="Doe")
    
    @type table: gluon.dal.Table
    
    @rtype: int
    @return: the C{id} of the record. 
    
    """
    id = get_id(table, **kwargs)
    
    if id == None:
        id = table.insert(**kwargs)
        
    return id


def get_field_validators(field):
    """Get the Ext JS configuration parameters to run 
    database gluon.dal.Field validators on the client-side.
    
    @type field: gluon.dal.Field
    
    @rtype: dict
    @return: The keys depend on the validators. 
    Typical one are C{minValue}, C{maxValue}, C{regex}, C{regexText},
    C{maxLength}, C{minLength}, ...
    
    """
    cfg = {}
    
    validators = as_list(field.requires)
        
    for el in validators:
    
        vdt = el
        if isinstance(vdt, IS_EMPTY_OR):
            vdt = vdt.other
            
        # date and time validators
        if isinstance(vdt, IS_DATE_IN_RANGE):
            format = '%Y-%m-%d'
            cfg['minValue'] = vdt.minimum.strftime(format)
            cfg['maxValue'] = vdt.maximum.strftime(format)

        elif isinstance(vdt, IS_DATETIME_IN_RANGE):
            format = '%Y-%m-%d %H:%M:%S'
            cfg['minValue'] = vdt.minimum.strftime(format)
            cfg['maxValue'] = vdt.maximum.strftime(format)

        # number validators
        elif isinstance(vdt, (IS_DECIMAL_IN_RANGE, IS_FLOAT_IN_RANGE, IS_INT_IN_RANGE)):
            cfg['minValue'] = vdt.minimum
            cfg['maxValue'] = vdt.maximum
        
        # string validator
        elif isinstance(vdt, IS_MATCH):
            cfg["regex"] = vdt.regex.pattern
            cfg["regexText"] = vdt.error_message
        
        # others validator
        elif isinstance(vdt, IS_LENGTH):
            cfg["maxLength"] = vdt.maxsize
            cfg["minLength"] = vdt.minsize
            
    return cfg


def get_file_paths(path, ext=None, alpha=True):
    """get the local paths for files matching the arguments.
    For example, to get the paths of the javascript source files
    located in the local directory C{static/plugin_dbui/src}::
    
        li = get_file_paths("static/plugin_dbui/src", ext=".js")

    The local paths are relative to the application directory.  
    
    @type path: None, list or string
    @param path: local path defining the target files:

        - C{None}.
        
        - a C{string} pointing to one files.

        - a C{list} of strings pointing to several files.
        
        - a C{string} pointing to a directory.
       
        - a C{list} of string pointing to files and directory.
    

    @type ext: str
    @param ext: select only files with the extension C{ext}, I{i.e} 
    C{.js}, C{.css}, ...

    @type alpha: bool
    @param alpha: sort the paths in alphabetic order
    
    @rtype: list
    @return: local paths for all selected files.
        
    """
    li = []
    if path == None:
        return li
    
    server_path, client_path = get_reference_paths()
    
    for el in as_list(path):
        p = os.path.join(server_path, el)
        
        if os.path.isfile(p):
            if ext and (not el.endswith(ext)): continue
            li.append(os.path.join(client_path, el))
            
        elif os.path.isdir(p):
            for file in os.listdir(p):
                pfile = os.path.join(server_path, el, file)
                if os.path.isfile(pfile):
                    if ext and (not pfile.endswith(ext)): continue
                    li.append(os.path.join(client_path, el, file))
    
    if alpha:
        li.sort()
        
    return li


def get_foreign_field(field):
    """Get a tuple with the reference table name, reference field name and
    reference primary key name, for the C{field}.
    
    @type field: gluon.dal.Field
    
    @rtype: tuple
    @return:
    
        - C{(k_tablename, k_fieldname, k_keyname)}
        
        - C{k_tablename} is the name of the reference table
        
        - C{k_fieldname} is the name of the reference field, I{i.e}
          C{table.field} S{->} C{k_table.k_field}
        
        - C{k_keyname} is the name of the primary key of the reference table.
          It is requires by the left join, 
          C{table.field == k_table.k_key}.
          In almost all case C{k_key} is equal to C{id}.
    
        - C{None} for standard field.
    
    """
    if not is_foreign_field(field):
        return None
    
    # foreign field validators IS_IN_BD contains all information 
    # on the parent key for the left join and on the parent field
    validators = as_list(field.requires)

    # Several syntax are allowed

    # IS_IN_DB is use.
    for vdt in validators:
        
        if isinstance(vdt, IS_IN_DB):
            # IS_IN_DB contains only the parent field. 
            # The key for the left join is the id key of the parent table
            if len(vdt.ks) == 1:
                return (vdt.ktable, vdt.ks[0], 'id')
            
            # IS_IN_DB contains the key for the join and the parent field 
            else:
                return (vdt.ktable, vdt.ks[0], vdt.ks[1])   
    
    # IS_IN_DB is not use. 
    # The parent key for the left join is the id key of the parent table
    # the parent field is the first field of the parent table which differs id
    i = field.type.index(' ')
    ktablename = field.type[i:].strip()
    
    for kfield in field._db[ktablename]:
        if kfield.name != 'id':
            return (ktablename, kfield.name, 'id')

    raise Exception


def get_id(table, **kwargs):
    """Get the C{id} of the record identified by a set of C{field=value} 
    keyword arguments. Return C{None} if the record does not exist.

    @type table: gluon.dal.Table
    
    @rtype: int or None
    @return: the C{id} of the record. 
    
     """
    query = None
    for k, v in kwargs.iteritems():
        el = table[k] == v

        if query:
            query = query & (el)
        else:
            query = el
   
    row =  table(query)
    if row:
        return row.id
           
    return None


def get_language():
    """Get the primary language of the application from the HTML
    language code, I{i.e} C{en}, C{fr}, ....
    
    @rtype: str
    @return:
    
    """
    lg = current.T.accepted_language
    
    # HTML tag for language: primary-dialect
    # Extract the primary language and ignore dialect
    m = re.match('([a-z]+)(\-([a-zA-Z]*))?', lg)
    if m:
        return m.group(1)

    return lg


def get_plugin_path(plugin_name):
    """Get the local path of the plugin identified by its name.
    
    @type plugin_name: str
    @param plugin_name: name of the plugin, I{i.e} C{"plugin_dbui"}.
    
    @rtype: str
    @return: 
        - the path of the first directory starting with C{plugin_name}.
        - the local path is relative to the application directory.
        - C{None} if the plugin is not found.
    
    """
    server_path, client_path = get_reference_paths()

    # look for the full name of the plugin directory
    p_static = os.path.join(server_path, 'static')
    for el in os.listdir(p_static):
        if os.path.isdir(os.path.join(p_static, el)):
            if el.startswith(plugin_name):
                return os.path.join(client_path, 'static', el)

    # Nothing found
    return None

    
def get_reference_paths():
    """Get a tuple with the absolute and local paths:
    
        - The absolute path is relative to the web2py directory.
          It is useful on the server-side
        - The local path is relative to the applications directory.
          It is useful on the client-side
          
    @rtype: tuple
    @return: C{("applications/myapp", "/myapp")}

    """
    server_path = os.path.join('applications', current.request.application)
    client_path = os.path.join(os.path.sep, current.request.application)
    
    return (server_path, client_path)


def get_set_field(field):
    """Get the list of the set data for the C{field} when defines.
    
    @type field: gluon.dal.Field
    @param field:
    
    @rtype: list
    
    """
    validators = as_list(field.requires)

    for vdt in validators:
        if isinstance(vdt, IS_IN_SET):
            return vdt.theset
        
    return []


def get_script_path(plugin):
    """Get the local path of the main javascript script to be 
    ran on the client-side. 

        >>> import plugin_dbui as dbui
        >>> plugin = plugins.dbui
        >>> path = dbui.get_script_path(plugin)

    The main script can be selected in several ways:
        - via the default script
        - via the URL calling the application
        - via the configuration parameter of the plugin_dbui
    
    B{Via the default value:}
        - the default script is located in C{static/plugin_dbui/main.js}
        
    B{Via the URL:}
        - the URL should contain the argument C{script=foo} or C{script=foo.js}
        - the script is search in C{app_script_dir}, one of the C{plugin_dbui}
          configuration parameters.
           
    B{Via plugin_dbui configuration:}
        - the script path is defined by the C{plugin_dbui}
          configuration parameters C{app_script}.
        
    @type plugin: plugin instance
    @param plugin: 
        
    @rtype: str
    @return: 
        - the local path of the script file
        - the local paths is relative to the application directory.
        
    """
    server_path, client_path = get_reference_paths()

    # script name specifies in the URL
    if "script" in current.request.vars:
        script = current.request.vars.script
        
        if not script.endswith(".js"):
            script = '%s.js' % script

        pdir = os.path.join(server_path, plugin.app_script_dir)
        
        if script not in os.listdir(pdir):
            raise HTTP(500, 'Request script "%s" does not exist !!!' % script)
        
        return os.path.join(client_path, plugin.app_script_dir, script)

    # default script for the application
    if plugin.app_script:
        return os.path.join(client_path, plugin.app_script)
        
    # default script for the plugin
    return os.path.join(client_path, plugin.dbui_script)


def get_store_id(name):
    """Build the store identifier using C{name}.
    
    @type name: str
    @rtype: str
    @return: C{nameStore}
    
    """
    return STOREID % name


def get_versions():
    """Get the versions of the application and plugins.
    
    @rtype: dict
    @return:
        - C{dbui} the version of the plugin  dbui
        - C{extjs} the version of the plugin extjs
        - C{mathjax} the version of the plugin mathjax
        - C{web2py} the version of the framework web2oy
        - C{myapp} the version of my application
    
    """
    server_path, client_path = get_reference_paths()

    # plugin dbui
    fn = os.path.join(server_path, 'static', 'plugin_dbui', 'src', 'base.js')
    
    fi = open(fn, 'rb')
    s = fi.read()
    fi.close()
    
    m = re.match(".+App.version = '([\w._-]*)';.*", s, re.DOTALL)
    dbui = m.group(1)

    # plugin extjs
    extjs = current.T('plugin not install')
    path = os.path.join(server_path, 'static', 'plugin_extjs')
    if os.path.exists(path):
        fn = os.path.join(server_path, 'static', 
                                       'plugin_extjs', 
                                       'src', 
                                       'ext-core', 
                                       'src', 
                                       'core', 
                                       'Ext.js')
        
        fi = open(fn, 'rb')
        s = fi.read()
        fi.close()
        
        m = re.match(r".+ version : '(\d+(\.\d+)*)',", s, re.DOTALL)
        if m:
            extjs = m.group(1)
    
    # plugin mathjax
    mathjax = current.T('plugin not install')
    path = os.path.join(server_path, 'static', 'plugin_mathjax')
    if os.path.exists(path):
        fn = os.path.join(server_path, 'static', 'plugin_mathjax', 'MathJax.js')
        
        fi = open(fn, 'rb')
        s = fi.read()
        fi.close()
        
        m = re.match(r'.+MathJax.version="(\d+(\.\d+)*)";', s, re.DOTALL)
        if m:
            mathjax = m.group(1)

    # web2py
    web2py = ''
    val = current.request.env.web2py_version
    if isinstance(val, str):
        m = re.match('\d+\.\d+\.\d+', val)
        if m:
            web2py = m.group()

    elif isinstance(val, (tuple, list)):
        li = [str(el) for el in val if isinstance(el, int)]
        web2py = '.'.join(li)
        
    # version of the application
    myapp = current.T("unknown")
    
    if os.path.exists(GIT):
        ref_path = os.getcwd()
        os.chdir(server_path)
    
        fi = tempfile.TemporaryFile()
        subprocess.call(["git", "describe", "--tags"], stdout=fi)
        fi.seek(0)
        myapp = fi.read().strip('\n')
        fi.close()
        
        os.chdir(ref_path)

    return dict(dbui=dbui, extjs=extjs, mathjax=mathjax, web2py=web2py, myapp=myapp)


def get_where_query(table):
    """Build the L{gluon.dal.Query} resolving foreign keys
    for the C{table}.
    
    @type table: gluon.dal.Table
    @param table:
    
    @rtype: gluon.dal.Query or None
    @return: C{None} when the table has no foreign key
    
    """
    db = table._db
    query = None

    for field in table:
        if not is_foreign_field(field):
            continue
        
        k_table, k_field, k_key = get_foreign_field(field)
        condition = field == db[k_table][k_key]
        
        if query:
            query = ((query) & (condition)) 
        else:
            query = condition
            
    return query

    
def is_foreign_field(field):
    """Return C{True} when the C{field} is a reference field.
    
    @type field: gluon.dal.Field
    @param field:
    
    @rtype: bool
    
    """
    return field.type.startswith("reference")


def is_mathjax():
    """Return C{True} when the plugin mathjax is installed.
    
    @rtype: bool
    @return:
    
    """
    return get_plugin_path('plugin_mathjax') != None


def is_set_field(field):
    """Return C{True} when the C{field} value 
    should matches those of a C{set}.
    
    @type field: gluon.dal.Field
    
    @rtype: bool
    
    """
    validators = as_list(field.requires)

    for vdt in validators:
        if isinstance(vdt, IS_IN_SET):
            return True
        
    return False


def is_table_with_foreign_fields(table):
    """Return C{True} when the database C{table} has at least
    one reference field.
    
    @type table: gluon.dal.Table
    
    @rtype: bool
    
    """
    for field in table:
        if is_foreign_field(field):
            return True
    return False


def rows_serializer(rows):
    """Serialize a L{gluon.dal.Rows} as a list of dictionary.
    It handle the general case where field belongs to different
    table and where function are applied.

    @type rows: gluon.dal.Rows
    @rtype: list
    @return: 
        - C{[{Table1Field1: val1,...}, {Table1Field1: val2,...}, ...]}
        - A dictionary per row.
        - Field are encoded as TableField.
        - Field with an applied function like COUNT(table.field) 
          are encoded as TableFieldCount.
    
    """
    # regular expressions
    # A field name can only contains [a_zA-Z0-9_] according to web2py
    table_field = re.compile(r'[\w_]+\.[\w_]+') 
    func_table_field = re.compile(r'([A-Z]+)\(([\w_]+)\.([\w_]+)\)')

    # Map between column name and TableField
    map = {}
    for col in rows.colnames:
        if not table_field.match(col):
            func, table, field = func_table_field.match(col).groups()
            map[col] = encode_field(table, field, func)
        else:
            table, field = col.split(".")
            map[col] = encode_field(table, field)

    # serialize
    li = []
    for row in rows:
        di = {}
        for col in rows.colnames:
            if not table_field.match(col):
                di[map[col]] = row._extra[col]
            else:
                (table, field) = col.split(".")
                if table in row:
                    di[map[col]] = row[table][field]
                else:
                    di[map[col]] = row[field]
        li.append(di)

    return li